/***************************************************************************
                          frameitem.h  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sun Aug 29 1999
    copyright            : (C) 2005 by Werner Stille
    email                : stille@uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef FRAMEITEM_H
#define FRAMEITEM_H

#include "kplitem.h"

template<class> class QPtrList;

/**
  * Frame item class. Class for frame items with tics, grid, labels.
  * @author Werner Stille
  */
class FrameItem : public KplItem
{
public:
  /** Constructor */
  FrameItem();
  /**
    * Constructor. Initializes corresponding to autoplot settings.
    * @param aut pointer to structure containing autoplot settings.
    */
  FrameItem(KplNamespace::AutoStruct* aut);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KSimpleConfig object containing item data.
    * @param aut pointer to structure containing autoplot settings.
    */
  FrameItem(KSimpleConfig* plo, KplNamespace::AutoStruct* aut);
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param active true for visible item.
    * @param logx true for logarithmic x axis.
    * @param logy true for logarithmic y axis.
    * @param ndigx number of digits for x axis.
    * @param ndigy number of digits for y axis.
    * @param mticx number of intervals between major tics of x axis.
    * @param mticy number of intervals between major tics of y axis.
    * @param gridmode 0 for frame only, 1 for axes with tics,
    * 2 for axes with tics and labels, 3 for grid, 4 for grid with labels.
    * @param iex logarithm of normalization factor for x axis.
    * @param iey logarithm of normalization factor for y axis.
    * @param colFrame color for frame and labels.
    * @param colGrid color for grids and tics.
    * @param x0 left page margin in centimeters.
    * @param w width in centimeters.
    * @param y0 bottom page margin in centimeters.
    * @param h height in centimeters.
    * @param xmin x value of left data window margin in user units.
    * @param xmax x value of right data window margin in user units.
    * @param ymin y value of bottom data window margin in user units.
    * @param ymax y value of top window data margin in user units.
    * @param xtic distance between major tics of x axis.
    * @param ytic distance between major tics of y axis.
    * @param relSize relative size of text, markers and lines.
    * @param sx x axis text.
    * @param sy y axis text.
    * @param sh heading text.
    * @param aut pointer to structure containing autoplot settings.
    * @param xOff offset for first major x tic.
    * @param yOff offset for first major y tic.
    * @param xBottom true for x axis labels on bottom side.
    * @param xTop true for x axis labels on top side.
    * @param yLeft true for y axis labels on left side.
    * @param yRight true for y axis labels on right side.
    * @param collapsed true for collapsed in list view.
    */
  FrameItem(bool active, bool logx, bool logy, int ndigx, int ndigy,
            int mticx, int mticy, int gridmode, int iex, int iey,
            const QString& colFrame, const QString& colGrid,
            double x0, double w, double y0, double h,
            double xmin, double xmax, double ymin, double ymax,
            double xtic, double ytic, double relSize,
            const QString& sx, const QString& sy, const QString& sh,
            KplNamespace::AutoStruct* aut, double xOff = 0.0,
            double yOff = 0.0, bool xBottom = true, bool xTop = false,
            bool yLeft = true, bool yRight = false, bool collapsed = false);
  /** Destructor */
  virtual ~FrameItem();
  virtual ItemTypes iType() const;
  virtual void draw(KplGraph* g);
#ifndef KPL_CLASSES_ONLY
  virtual void writePlo(KSimpleConfig* plo, const KURL& url, bool abs,
                        KplDoc* m) const;
  virtual void setText(QListViewItem* it, bool* arrays, bool* funcs) const;
  virtual int editItem(QWidget* parent, KplDoc* m, int i);
#endif
  virtual KplItem* copy() const;
  /**
   * Sets collapse state for list view.
   * @param b state.
   */
  inline void setCollapsed(bool b) { collapsed = b; };
  /**
   * Returns clloapse state for list view.
   * @return state.
   */
  inline bool isCollapsed() const { return collapsed; };
  /**
   * Calculates scaling parameters for a frame item and all its child items.
   * @param autoNorm true for automatic normalization.
   * @param item index of frame item.
   * @param items list of plot items.
   * @param fx normalization factor for x axis.
   * @param fy normalization factor for y axis.
   * @return true when scaling was successfull.
   */
  bool calcAutoScale(bool autoNorm, int item, QPtrList<KplItem>* items,
                     double* fx, double* fy);
  /**
   * Normalizes all child items of a frame item.
   * @param item index of frame item.
   * @param items list of plot items.
   * @param fx normalization factor for x axis.
   * @param fy normalization factor for y axis.
   */
  void normalize(int item, QPtrList<KplItem>* items, double fx, double fy);
  /**
   * Calculates and sets the scaling parameters.
   * @param xmin minimum x value.
   * @param xmax maximum x value.
   * @param fx x normalization factor.
   * @param ymin minimum y value.
   * @param ymax maximum y value.
   * @param fy y normalization factor.
   */
  void autoScale(double xmin, double xmax, double fx, double ymin, double ymax,
                 double fy);
  /**
   * Scales automatically the frame item and all its child items.
   * @param autoNorm true for automatic normalization.
   * @param items list of plot items.
   * @return true when scaling was successfull.
   */
  bool autoScale(bool autoNorm, QPtrList<KplItem>* items);

protected:
  bool logx, logy, xBottom, xTop, yLeft, yRight, collapsed;
  int ndigx, ndigy, mticx, mticy, gridmode, iex, iey;
  unsigned colFrame, colGrid;
  double xf, yf, w, h, xmi, xma, ymi, yma, xtic, ytic, xOff, yOff;
  QString sx, sy, sh;

  friend class FrameDlg;
  friend class KplDoc;
  friend class KplView;
};

#endif
